// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "games/vn/components/VNManager.hpp"
#include "scene/SceneItem.hpp"

namespace Dawn {
  class VNEvent : public StateOwner {
    public:
      StateEvent<> eventFinished;

      /**
       * Invoked by a parent VN Manager. This is the start of the event.
       * 
       * @param parent The VN Event Parent. Usually the manager but not always.
       * @param previous The previous event that was running before this one.
       */
      void start(
        IVNEventParent *parent,
        VNEvent *previous
      );

    protected:
      VNManager *manager = nullptr;
      IVNEventParent *parent = nullptr;
      VNEvent *doNext = nullptr;
      VNEvent *previous = nullptr;
      bool_t finished = false;

      /**
       * Initializes the event. This is called by the VNManager, and should not
       * be called by anything else.
       * 
       * @param manager The VNManager that is running this event.
       */
      void init(VNManager *manager);

      /**
       * Invoked by the VNManager, this is the end of the event. Perform the 
       * necessary cleanup, but remember that events may be re-started again
       * later.
       */
      void end();

      /**
       * Overrideable method that is called when the event is started.
       */
      virtual void onStart();

      /**
       * Overrideable method that is called when the event is ended.
       */
      virtual void onEnd();

    public:
      /**
       * Returns the scene this event is running in.
       * @return Pointer to the scene.
       */
      Scene * getScene();

      /**
       * End this event and move on to the next event.
       */
      void next();

      /**
       * Returns the next event to be executed after this one. Can be overridden
       * to return a different event other than the doNext event.
       * @return Pointer to the next event.
       */
      virtual VNEvent * getNextEvent();

      /**
       * Chains an event to be executed after this event has finished.
       * 
       * @param next Event to process next.
       * @return Whatever you pass in to next.
       */
      template<class T>
      T * then(T *next) {
        this->doNext = next;
        return next;
      }

      friend class VNManager;
  };
}